import django
from django.conf import global_settings, settings

settings.configure(
    # Minimal django settings
    DATABASES={
        'default': {
            'ENGINE': 'django.db.backends.sqlite3',
            'NAME': 'unittest.sqlite',
            'TEST_NAME': ':memory:',
        },
    },
    MIDDLEWARE_CLASSES=global_settings.MIDDLEWARE_CLASSES,
    ROOT_URLCONF=None,
    INSTALLED_APPS=['etcd_settings'],
    DJES_DEV_PARAMS=None,
    DJES_REQUEST_GETTER=None,
    DJES_ENV='dev',
    DJES_ETCD_DETAILS=None,
    DJES_WSGI_FILE=None
    # DJES_ETCD_DETAILS=dict(host='localhost', port=4000, protocol='http',
    #                        prefix='/config/my-django-app')
)
if hasattr(django, 'setup'):
    django.setup()



class TestSettings(object):

    ETCD_PREFIX = '/config/etcd_settings'
    ETCD_ENV = 'test'
    ETCD_HOST = 'etcd'
    ETCD_PORT = 2379
    ETCD_USERNAME = 'test'
    ETCD_PASSWORD = 'test'
    ETCD_DETAILS = dict(
        host='etcd',
        port=2379,
        prefix='/config/etcd_settings',
        username='test',
        password='test'
    )

settings = TestSettings()



import json
import logging
import re
import time
from importlib import import_module
from os import utime

from etcd import Client, EtcdException, EtcdKeyNotFound

from .utils import (
    CustomJSONEncoder, attrs_to_dir, byteify, custom_json_decoder_hook,
    threaded,
)


class EtcdConfigInvalidValueError(Exception):
    def __init__(self, key, raw_value, value_error):
        self.key = key
        self.raw_value = raw_value
        self.value_error = value_error
        super(EtcdConfigInvalidValueError, self).__init__(
            "Invalid value for key '{}'. Raising '{}', because of value: '{}'"
            .format(key, value_error, raw_value))


class EtcdClusterState(object):
    etcd_index = 0


class EtcdConfigManager(object):
    def __init__(
            self, dev_params=None, prefix='config', protocol='http',
            host='localhost', port=2379, username=None, password=None,
            long_polling_timeout=50, long_polling_safety_delay=5):
        self._client = Client(
            host=host, port=port, protocol=protocol, allow_redirect=True,
            username=username, password=password)
        # Overriding retries for urllib3.PoolManager.connection_pool_kw
        self._client.http.connection_pool_kw['retries'] = 0
        self._base_config_path = prefix
        self._dev_params = dev_params
        self._base_config_set_path = "{}/extensions" \
            .format(self._base_config_path)
        r = ('^(?P<path>{}/(?:extensions/)?'
             '(?P<envorset>[\w\-\.]+))/(?P<key>.+)$')
        self._key_regex = re.compile(r.format(self._base_config_path))
        self.long_polling_timeout = long_polling_timeout
        self.long_polling_safety_delay = long_polling_safety_delay
        self._init_logger()

    def _init_logger(self):
        self.logger = logging.getLogger('etcd_config_manager')
        logger_console_handler = logging.StreamHandler()
        logger_console_handler.setLevel(logging.ERROR)
        self.logger.addHandler(logger_console_handler)

    def _env_defaults_path(self, env='test'):
        return "{}/{}".format(self._base_config_path, env)

    def _config_set_path(self, set_name):
        return "{}/{}".format(self._base_config_set_path, set_name)

    def _encode_config_key(self, k):
        return k.lower().replace('_', '/')

    def _decode_config_key(self, k):
        [env_or_set, key_path] = re.sub(
            self._key_regex, '\g<envorset>|\g<key>', k).split('|')
        return env_or_set, key_path.upper().replace('/', '_')

    def _encode_config_value(self, val):
        return json.dumps(val, cls=CustomJSONEncoder)

    def _decode_config_value(self, val):
        decoded = json.loads(val, object_hook=custom_json_decoder_hook)
        return byteify(decoded)

    def _process_response_set(self, rset, env_defaults=True):
        d = {}
        EtcdClusterState.etcd_index = rset.etcd_index
        for leaf in rset.leaves:
            try:
                config_set, key = self._decode_config_key(leaf.key)
            except ValueError:
                info = "An error occurred when processing an EtcdResponse"
                if not env_defaults:
                    info += " (is '{}' a directory?)".format(
                        self._base_config_set_path)
                self.logger.warning(info)
            else:
                if leaf.value is not None:
                    try:
                        value = self._decode_config_value(leaf.value)
                    except ValueError as e:
                        raise EtcdConfigInvalidValueError(
                            leaf.key, leaf.value, e)

                    if env_defaults:
                        d[key] = value
                    else:
                        if config_set not in d:
                            d[config_set] = {}
                        d[config_set][key] = value
        return d

    @staticmethod
    def get_dev_params(mod):
        params = {}
        if mod:
            params = attrs_to_dir(import_module(mod))
        return params

    def get_env_defaults(self, env):
        res = self._client.read(
            self._env_defaults_path(env),
            recursive=True)
        conf = self._process_response_set(res)
        conf.update(EtcdConfigManager.get_dev_params(self._dev_params))
        return conf

    def get_config_sets(self):
        conf = {}
        try:
            res = self._client.read(
                self._base_config_set_path,
                recursive=True)
            conf = self._process_response_set(res, env_defaults=False)
        except EtcdKeyNotFound:
            self.logger.warning(
                "Unable to find config sets at '{}' (expected a dict)",
                self._base_config_set_path)
        return conf

    @threaded(daemon=True)
    def monitor_env_defaults(
            self, env, conf={}, wsgi_file=None, max_events=None):
        processed_events = 0
        for event in self._watch(
                self._env_defaults_path(env), conf, wsgi_file, max_events):
            if event is not None:
                conf.update(self._process_response_set(event))
                conf.update(EtcdConfigManager.get_dev_params(self._dev_params))
                if wsgi_file:
                    with open(wsgi_file, 'a'):
                        utime(wsgi_file, None)
            processed_events += 1
        return processed_events

    @threaded(daemon=True)
    def monitor_config_sets(self, conf={}, max_events=None):
        processed_events = 0
        for event in self._watch(
                self._base_config_set_path, conf=conf, max_events=max_events):
            if event is not None:
                conf.update(
                    self._process_response_set(event, env_defaults=False))
            processed_events += 1
        return processed_events

    def _watch(self, path, conf={}, wsgi_file=None, max_events=None):
        i = 0
        while (max_events is None) or (i < max_events):
            try:
                i += 1
                index = EtcdClusterState.etcd_index
                if index > 0:
                    index = index + 1
                    res = self._client.watch(
                        path,
                        index=index,
                        recursive=True,
                        timeout=self.long_polling_timeout)
                else:
                    res = self._client.read(path, recursive=True)
                yield res
            except Exception as e:
                if not (isinstance(e, EtcdException)
                        and ('timed out' in str(e))):
                    self.logger.error("Long Polling Error: {}".format(e))
                    time.sleep(self.long_polling_safety_delay)
                yield None

    def set_env_defaults(self, env, conf={}):
        path = self._env_defaults_path(env)
        errors = {}
        for k, v in conf.items():
            if k.isupper():
                try:
                    encoded_key = self._encode_config_key(k)
                    self._client.write(
                        "{}/{}".format(path, encoded_key),
                        self._encode_config_value(v))
                except Exception as e:
                    errors[k] = str(e)
        return errors

    def set_config_sets(self, config_sets={}):
        errors = {}
        for set_name, config_set in config_sets.items():
            path = self._config_set_path(set_name)
            for k, v in config_set.items():
                if k.isupper():
                    try:
                        self._client.write(
                            "{}/{}".format(path, self._encode_config_key(k)),
                            self._encode_config_value(v))
                    except Exception as e:
                        errors[k] = str(e)
        return errors



import logging
import unittest

from etcd_settings import utils


class TestLoggingFilter(unittest.TestCase):

    def setUp(self):
        self.logger_filter = utils.IgnoreMaxEtcdRetries()

    def test_log_record_without_args(self):
        self.assertTrue(
            self.logger_filter.filter(
                logging.LogRecord(
                    'etcd.client',
                    logging.ERROR,
                    '/',
                    0,
                    'Read timed out',
                    tuple(),
                    None
                )
            )
        )

    def test_log_record_match(self):
        self.assertFalse(
            self.logger_filter.filter(
                logging.LogRecord(
                    'etcd.client',
                    logging.ERROR,
                    '/',
                    0,
                    'Error %s exception %s',
                    ('Read timed out', 'MaxRetryError'),
                    None
                )
            )
        )



import os
import re
from importlib import import_module

from django.conf import settings as django_settings

from .manager import EtcdConfigManager
from .utils import (
    attrs_to_dir, copy_if_mutable, dict_rec_update, find_project_root,
)


class EtcdSettingsProxy(object):

    def __init__(self):
        self.env = getattr(django_settings, 'DJES_ENV', None)
        dev_params = getattr(django_settings, 'DJES_DEV_PARAMS', None)
        etcd_details = getattr(django_settings, 'DJES_ETCD_DETAILS', None)
        self._init_req_getter(
            getattr(django_settings, 'DJES_REQUEST_GETTER', None))
        self._locate_wsgi_file(
            getattr(django_settings, 'DJES_WSGI_FILE', None))
        if etcd_details is not None:
            self._etcd_mgr = EtcdConfigManager(dev_params, **etcd_details)
            self._config_sets = self._etcd_mgr.get_config_sets()
            self._env_defaults = self._etcd_mgr.get_env_defaults(self.env)
        else:
            self._etcd_mgr = None
            self._config_sets = dict()
            self._env_defaults = EtcdConfigManager.get_dev_params(dev_params)

    def _locate_wsgi_file(self, wsgi_file):
        if wsgi_file is None:
            self._wsgi_file = None
        elif wsgi_file.startswith(os.path.sep):
            self._wsgi_file = wsgi_file
        else:
            self._wsgi_file = os.path.join(
                find_project_root('manage.py'),
                wsgi_file)

    def _init_req_getter(self, s):
        if s is not None:
            r = re.compile('(?P<module>.*)\.(?P<f>[\w_]+)')
            m = re.match(r, s)
            mod_s = m.group('module')
            fun_s = m.group('f')
            mod = import_module(mod_s)
            self._req_getter = getattr(mod, fun_s)
        else:
            self._req_getter = None

    def _parse_req_config_sets(self):
        sets = []
        if self._req_getter is not None:
            request = self._req_getter()
            if request and getattr(request, "META", None):
                sets = request.META.get('HTTP_X_DYNAMIC_SETTING', '').split()
        return sets

    def start_monitors(self):
        if self._etcd_mgr is not None:
            self._etcd_mgr.monitor_env_defaults(
                env=self.env, conf=self._env_defaults,
                wsgi_file=self._wsgi_file)
            self._etcd_mgr.monitor_config_sets(conf=self._config_sets)

    def __getattr__(self, attr):
        try:
            dj_value = getattr(django_settings, attr)
            dj_value_exists = True
        except AttributeError:
            dj_value_exists = False
            dj_value = None
        try:
            value = self._env_defaults[attr]
            value_exists = True
        except KeyError:
            value_exists = dj_value_exists
            value = dj_value

        for override_set in self._parse_req_config_sets():
            config_set = self._config_sets.get(override_set, {})
            if attr in config_set:
                new_value = config_set[attr]
                value = copy_if_mutable(value)
                if isinstance(value, dict) and isinstance(new_value, dict):
                    dict_rec_update(value, new_value)
                else:
                    value = new_value
        if value or value_exists:
            return value
        else:
            raise AttributeError(attr)

    def as_dict(self):
        items = attrs_to_dir(django_settings)
        items.update(self._env_defaults)
        return items


proxy = EtcdSettingsProxy()



import json
import os
import re
import sys

from django.http import HttpRequest
from django.test import TestCase
from django.test.utils import override_settings
from etcd_settings.loader import get_overwrites
from etcd_settings.manager import EtcdClusterState, EtcdConfigManager
from etcd_settings.proxy import EtcdSettingsProxy
from mock import MagicMock

from .conftest import settings


@override_settings(
    DJES_ETCD_DETAILS=settings.ETCD_DETAILS,
    DJES_ENV=settings.ETCD_ENV,
    DJES_REQUEST_GETTER='etcd_settings.utils.threaded',
    E=0
)
class TestEtcdSettingsProxy(TestCase):

    def setUp(self):
        s = ''
        with open('tests/unicode.txt', 'rb+') as f:
            if sys.version_info.major == 3:
                s = f.read().decode()
            else:
                s = f.read()
        self.mgr = EtcdConfigManager(
            prefix=settings.ETCD_PREFIX, host=settings.ETCD_HOST,
            port=settings.ETCD_PORT,
            username=settings.ETCD_USERNAME, password=settings.ETCD_PASSWORD
        )

        self.env_config = {
            "A": 1, "B": "c", "D": {"e": "f"}, "E": 1,
            "C": {'c2': 1}, 'ENCODING': s
        }
        self.mgr.set_env_defaults('test', self.env_config)
        self.mgr.set_config_sets({
            'foo': {'A': 11},
            'bar': {'C': {'c3': 2}}})
        self.proxy = EtcdSettingsProxy()
        with open('manage.py', 'w') as f:
            f.write("testing artifact")

    def tearDown(self):
        try:
            os.remove('manage.py')
        except:
            pass

    def test_loader_etcd_index_in_manager(self):
        env = get_overwrites(settings.ETCD_ENV, None, settings.ETCD_DETAILS)
        self.assertIsNotNone(env)
        self.assertGreater(EtcdClusterState.etcd_index, 0)

    def test_username_password(self):
        self.assertEquals({'authorization': u'Basic dGVzdDp0ZXN0'},
                          self.mgr._client._get_headers())

    def test_proxy_starts_without_extensions(self):
        self.mgr._client.delete(self.mgr._base_config_set_path, recursive=True)
        p = EtcdSettingsProxy()
        self.assertIsNotNone(p)

    def test_proxy_starts_when_extensions_is_not_a_dir(self):
        self.mgr._client.delete(self.mgr._base_config_set_path, recursive=True)
        self.mgr._client.write(
            self.mgr._base_config_set_path,
            json.dumps('not_a_dict'))
        p = EtcdSettingsProxy()
        self.assertIsNotNone(p)

    def test_proxy_reads_initial_blob(self):
        self.assertEquals(1, self.proxy.A)
        self.assertEquals("c", self.proxy.B)

    def test_proxy_raises_attribute_errors_on_not_found(self):
        with self.assertRaises(AttributeError):
            self.proxy.KEY_THAT_IS_NOT_THERE

    def test_proxy_reads_django_settings(self):
        self.assertEquals('test', self.proxy.DJES_ENV)

    def test_proxy_gives_prio_to_env_over_django_settings(self):
        self.assertEquals(1, self.proxy.E)

    def test_proxy_can_be_viewed_as_dict(self):
        d = self.proxy.as_dict()
        for k, v in self.env_config.items():
            self.assertEqual(v, d[k])

    def test_proxy_uses_dynamic_settings(self):
        r = HttpRequest()
        r.META = {'HTTP_X_DYNAMIC_SETTING': 'foo'}
        self.proxy._req_getter = MagicMock(return_value=r)
        self.assertEqual(11, self.proxy.A)

    def test_proxy_dynamic_settings_handle_dict_overwrites(self):
        r = HttpRequest()
        r.META = {'HTTP_X_DYNAMIC_SETTING': 'bar'}
        self.proxy._req_getter = MagicMock(return_value=r)
        c = self.proxy.C
        self.assertEqual(1, c.get('c2'))
        self.assertEqual(2, c.get('c3'))

    def test_proxy_locates_uwsgi_file(self):
        self.proxy._locate_wsgi_file(None)
        self.assertEqual(None, self.proxy._wsgi_file)
        self.proxy._locate_wsgi_file(__file__)
        self.assertEqual(__file__, self.proxy._wsgi_file)
        self.proxy._locate_wsgi_file('etcd_settings/proxy.py')
        self.assertIsNotNone(
            re.match("^/(.*)/etcd_settings/proxy.py", self.proxy._wsgi_file))
        os.remove('manage.py')
        with self.assertRaises(IOError):
            self.proxy._locate_wsgi_file('file_that_cannot_exist.py')

    def test_loader_gets_overwrites(self):
        self.assertEqual(
            self.env_config,
            get_overwrites(settings.ETCD_ENV, None, settings.ETCD_DETAILS)
        )



from .manager import EtcdConfigManager


def get_overwrites(env, dev_params, etcd_details):
    overwrites = {}
    if etcd_details is not None:
        mgr = EtcdConfigManager(dev_params, **etcd_details)
        overwrites = mgr.get_env_defaults(env)
    else:
        overwrites = EtcdConfigManager.get_dev_params(dev_params)
    return overwrites



import datetime
import json
import logging
import os
import time

from django.test import TestCase
from etcd import EtcdKeyNotFound
from etcd_settings.manager import (
    EtcdClusterState, EtcdConfigInvalidValueError, EtcdConfigManager,
)

from .conftest import settings


class TestEtcdConfigManager(TestCase):
    longMessage = True

    def _dataset_with_empty_dir(self):
        key = os.path.join(self.mgr._env_defaults_path(self.env), 'dir/empty')
        self.mgr._client.write(key, None, dir=True)
        value = self.mgr._client.get(key)
        return key, value

    def _dataset_with_invalid_json(self):
        key = os.path.join(
            self.mgr._env_defaults_path(self.env), 'json/invalid')
        self.mgr._client.set(key, '{')
        value = self.mgr._client.get(key)
        return key, value

    def _dataset_for_defaults(self):
        dataset = {}
        k1 = os.path.join(self.mgr._env_defaults_path(self.env), 'foo/bar')
        dataset[k1] = '"baz"'
        k2 = os.path.join(self.mgr._env_defaults_path(self.env), 'foo/baz')
        dataset[k2] = '"bar"'
        k3 = os.path.join(self.mgr._env_defaults_path(self.env), 'foobarbaz')
        dataset[k3] = '"superbaz"'
        expected_env = {
            'FOO_BAR': 'baz',
            'FOO_BAZ': 'bar',
            'FOOBARBAZ': 'superbaz',
        }
        for k, v in dataset.items():
            self.mgr._client.set(k, v)
        return dataset.keys(), expected_env

    def _dataset_for_configsets(self):
        dataset = {}
        k1 = os.path.join(self.mgr._config_set_path('foo'), 'bar')
        dataset[k1] = '1'
        k2 = os.path.join(self.mgr._config_set_path('foo'), 'baz')
        dataset[k2] = '2'
        k3 = os.path.join(self.mgr._config_set_path('foo.bar'), 'baz')
        dataset[k3] = '1'
        k4 = os.path.join(self.mgr._config_set_path('foo.bar'), 'bazbaz')
        dataset[k4] = '2'
        k5 = os.path.join(self.mgr._config_set_path('foo.bar-zoo'), 'bar')
        dataset[k5] = '1'
        expected_sets = {
            'foo': {'BAR': 1, 'BAZ': 2},
            'foo.bar': {'BAZ': 1, 'BAZBAZ': 2},
            'foo.bar-zoo': {'BAR': 1},
        }
        for k, v in dataset.items():
            self.mgr._client.set(k, v)
        return dataset.keys(), expected_sets

    def setUp(self):

        self.env = 'unittest'
        EtcdClusterState.etcd_index = 0
        self.mgr = EtcdConfigManager(
            dev_params=None, prefix=settings.ETCD_PREFIX, protocol='http',
            host=settings.ETCD_HOST, port=settings.ETCD_PORT,
            username=settings.ETCD_USERNAME, password=settings.ETCD_PASSWORD,
            long_polling_timeout=0.1, long_polling_safety_delay=0.1
        )
        for l in self.mgr.logger.handlers:
            l.setLevel(logging.CRITICAL)

    def tearDown(self):
        def try_unless_not_found(f):
            try:
                f()
            except EtcdKeyNotFound:
                pass

        def clean_env():
            self.mgr._client.delete(
                self.mgr._env_defaults_path(self.env), recursive=True)

        def clean_extensions():
            self.mgr._client.delete(
                self.mgr._base_config_set_path, recursive=True)

        try_unless_not_found(clean_extensions)
        try_unless_not_found(clean_env)

    def test_init_logger(self):
        self.assertIsNotNone(self.mgr.logger)

    def test_encode_config_key(self):
        self.assertEqual(
            'foo/bar/baz',
            self.mgr._encode_config_key('FOO_BAR_BAZ'))

    def test_decode_env_config_key(self):
        key = 'FOO_BAR'
        s = '{}/{}/foo/bar'.format(self.mgr._base_config_path, self.env)
        self.assertEqual((self.env, key), self.mgr._decode_config_key(s))

    def test_decode_set_config_key(self):
        key = 'FOO_BAR'
        configset = 'unit.test'
        s = '{}/{}/foo/bar'.format(self.mgr._base_config_set_path, configset)
        self.assertEqual((configset, key), self.mgr._decode_config_key(s))

    def test_encode_config_value(self):
        self.assertEqual(
            '"abcde"',
            self.mgr._encode_config_value('abcde'))
        self.assertEqual(
            '112',
            self.mgr._encode_config_value(112))
        self.assertEqual(
            # Tuples are lost in encoding, should be avoided as config values
            '[1, "b"]',
            self.mgr._encode_config_value((1, 'b')))
        encoded_config = self.mgr._encode_config_value(dict(foo=1, bar='baz'))
        self.assertEqual(
            json.loads('{"foo": 1, "bar": "baz"}'),
            json.loads(encoded_config))

    def test_process_response_set_empty(self):
        key, value = self._dataset_with_empty_dir()
        self.assertEqual({}, self.mgr._process_response_set(value))

    def test_process_response_exception_handling(self):
        with self.assertRaises(EtcdConfigInvalidValueError) as excContext:
            key, value = self._dataset_with_invalid_json()
            self.mgr._process_response_set(value)

        exc = excContext.exception

        self.assertEqual(key, exc.key)
        self.assertEqual(value.value, exc.raw_value)
        self.assertIn(key, str(exc), "Expect key in message")
        self.assertIn(
            "Expecting", str(exc), msg="Expect detailed error message")
        self.assertIn(
            "line", str(exc), msg="Expect line number in error message")
        self.assertIn(
            "column", str(exc), msg="Expect column number in error message")
        self.assertIn(value.value, str(exc), msg="Expect invalid value")

    def test_decode_config_value(self):
        self.assertEqual(
            'abcde',
            self.mgr._decode_config_value('"abcde"'))
        self.assertEqual(
            112,
            self.mgr._decode_config_value('112'))
        self.assertEqual(
            dict(foo=1, bar='baz'),
            self.mgr._decode_config_value('{"foo": 1, "bar": "baz"}'))
        self.assertEqual(
            [1, 'b'],
            self.mgr._decode_config_value('[1, "b"]'))

    def test_custom_encoding_decoding_values(self):
        d = datetime.datetime(2015, 10, 9, 8, 7, 6)
        encoded_d = self.mgr._encode_config_value(d)
        decoded_d = self.mgr._decode_config_value(encoded_d)
        self.assertEqual(True, isinstance(decoded_d, datetime.datetime))
        self.assertEqual(d.isoformat(), decoded_d.isoformat())
        self.assertEqual(d, decoded_d)

    def test_get_env_defaults(self):
        keys, expected = self._dataset_for_defaults()
        self.assertEqual(expected, self.mgr.get_env_defaults(self.env))

    def test_get_config_sets(self):
        keys, expected_sets = self._dataset_for_configsets()
        self.assertEqual(expected_sets, self.mgr.get_config_sets())

    def test_monitor_env_defaults(self):
        keys, expected_env = self._dataset_for_defaults()
        d = {}
        old_etcd_index = EtcdClusterState.etcd_index
        t = self.mgr.monitor_env_defaults(env=self.env, conf=d, max_events=1)
        self.assertEqual(1, t.result)
        self.assertEqual(expected_env, d)
        self.assertGreater(EtcdClusterState.etcd_index, old_etcd_index)

    def test_monitor_config_sets(self):
        keys, expected_sets = self._dataset_for_configsets()
        d = {}
        old_etcd_index = EtcdClusterState.etcd_index
        t = self.mgr.monitor_config_sets(conf=d, max_events=1)
        self.assertEqual(1, t.result)
        self.assertEqual(expected_sets, d)
        self.assertGreater(EtcdClusterState.etcd_index, old_etcd_index)

    def test_monitors_delay_and_continue_on_exception(self):
        d = {}
        max_events = 2
        lambdas = [
            lambda: self.mgr.monitor_env_defaults(
                self.env, conf=d, max_events=max_events),
            lambda: self.mgr.monitor_config_sets(
                conf=d, max_events=max_events),
        ]
        try:
            self.mgr._base_config_path = 'Unknown'
            for l in lambdas:
                t0 = time.time()
                t = l()
                self.assertEqual(max_events, t.result)
                self.assertEqual(
                    True,
                    (time.time() - t0)
                    > (max_events * self.mgr.long_polling_safety_delay)
                )
                self.assertEqual(False, t.is_alive())
        finally:
            self.mgr._base_config_path = settings.ETCD_PREFIX

    def test_monitors_continue_on_etcd_exception(self):
        d = {}
        max_events = 2
        lambdas = [
            lambda: self.mgr.monitor_env_defaults(
                self.env, conf=d, max_events=max_events),
            lambda: self.mgr.monitor_config_sets(
                conf=d, max_events=max_events),
        ]
        try:
            old_etcd_index = EtcdClusterState.etcd_index
            self.mgr._etcd_index = 999990
            for l in lambdas:
                t0 = time.time()
                t = l()
                self.assertEqual(max_events, t.result)
                net_time = time.time() - t0
                self.assertGreater(
                    net_time,
                    (max_events * self.mgr.long_polling_timeout)
                )
                self.assertLess(
                    net_time,
                    (max_events *
                        (self.mgr.long_polling_safety_delay
                            + self.mgr.long_polling_timeout))
                )
                self.assertEqual(False, t.is_alive())
        finally:
            EtcdClusterState.etcd_index = old_etcd_index



from django.test import TestCase
from etcd_settings.loader import get_overwrites


class TestLoader(TestCase):

    def test_get_overwrites_without_etcd(self):
        overwrites = get_overwrites(
            'unittest', 'tests.loader_dev_params', None)
        self.assertEqual('bar', overwrites['FOO'])



import copy
import datetime
import json
import logging
import os
import sys
from collections import Mapping
from functools import wraps
from threading import Thread

import six
from dateutil.parser import parse as parse_date


def attrs_to_dir(mod):
    data = {}
    for attr in dir(mod):
        if attr == attr.upper():
            data[attr] = getattr(mod, attr)
    return data


def dict_rec_update(d, u):
    """Nested update of a dict, handy for overriding settings"""
    # https://stackoverflow.com/questions/3232943/update-value-of-a-nested-dictionary-of-varying-depth
    for k, v in u.items():
        if isinstance(v, Mapping):
            r = dict_rec_update(d.get(k, {}), v)
            d[k] = r
        else:
            d[k] = u[k]
    return d


class Task(Thread):
    """
    The Threaded object returned by the @threaded decorator below
    """

    def __init__(self, method, *args, **kwargs):
        super(Task, self).__init__()
        self.method = method
        self.args = args
        self.kwargs = kwargs
        self._result = None
        self.__exc_info = None

    def run(self):
        try:
            self._result = self.method(*self.args, **self.kwargs)
        except:
            self.__exc_info = sys.exc_info()

    @property
    def result(self):
        self.join()
        if self.__exc_info is not None:
            six.reraise(*self.__exc_info)
        return self._result


def threaded(function=None, daemon=False):

    def wrapper_factory(func):

        @wraps(func)
        def get_thread(*args, **kwargs):
            t = Task(func, *args, **kwargs)
            if daemon:
                t.daemon = True
            t.start()
            return t

        return get_thread

    if function:
        return wrapper_factory(function)
    else:
        return wrapper_factory


class CustomJSONEncoder(json.JSONEncoder):

    custom_type_key = '_custom_type'
    custom_type_value_key = 'value'
    DECODERS = {'datetime': parse_date}

    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            return {self.custom_type_key: 'datetime',
                    self.custom_type_value_key: obj.isoformat()}
        else:
            return super(CustomJSONEncoder, self).default(obj)


def custom_json_decoder_hook(obj):

    ct = obj.get(CustomJSONEncoder.custom_type_key, None)
    if ct is not None:
        value = obj.get(CustomJSONEncoder.custom_type_value_key)
        return CustomJSONEncoder.DECODERS[ct](value)
    else:
        return obj


def find_project_root(root_indicator='manage.py', current=os.getcwd()):
    parent = os.path.dirname(current)
    if root_indicator in os.listdir(current):
        return current
    elif current == parent:
        # We are at '/' already!
        raise IOError('Not found: {}'.format(root_indicator))
    else:
        return find_project_root(root_indicator, parent)


def copy_if_mutable(value):
    """
    Copy function handling mutable values (only dicts and lists).
    """
    if type(value) in (dict, list):
        return copy.deepcopy(value)
    return value


def byteify(input):
    if isinstance(input, dict):
        return {byteify(key): byteify(value) for key, value in input.items()}
    elif isinstance(input, list):
        return [byteify(element) for element in input]
    elif (sys.version_info.major == 2) and (isinstance(input, unicode)):
        return input.encode('utf-8')
    else:
        return input


class IgnoreMaxEtcdRetries(logging.Filter):
    """
    Skip etcd.client MaxRetryError on timeout
    """

    def __init__(self, name='etcd.client'):
        super(IgnoreMaxEtcdRetries, self).__init__(name)

    def filter(self, record):
        msg = '{}'.format(record.args)
        return not (
            'MaxRetryError' in msg and
            'Read timed out' in msg
        )



from django.apps import AppConfig
from etcd_settings.proxy import proxy


class EtcdMonitor(AppConfig):

    name = 'etcd_settings'
    verbose_name = 'EtcdMonitor starting the threads for long polling'

    def ready(self):
        proxy.start_monitors()
